""" TemplateXslRendering API calls
"""
from core_main_app.components.template_xsl_rendering.models import TemplateXslRendering
from core_main_app.components.template import api as template_api
from core_main_app.commons import exceptions


def add_or_delete(template_id, list_xslt, detail_xslt, template_xsl_rendering_id=None):
    """ Manage the saving of a TemplateXslRendering. If no XSLTs have been given, deletes the instance.

    Args:
        template_id: TemplateXslRendering.
        list_xslt: XSLT.
        detail_xslt: XSLT.
        template_xsl_rendering_id: Instance id.

    Returns:

    """
    # Boolean to know if we need this instance in database, i.e there are XSLTs information.
    need_to_be_kept = list_xslt is not None or detail_xslt is not None

    try:
        # Update the configuration.
        template_xsl_rendering = get_by_id(template_xsl_rendering_id)
        if need_to_be_kept:
            template_xsl_rendering.list_xslt = list_xslt
            template_xsl_rendering.detail_xslt = detail_xslt
            _upsert(template_xsl_rendering)
        else:
            delete(template_xsl_rendering)
    except (Exception, exceptions.DoesNotExist):
        # If no configuration, create a new one.
        if need_to_be_kept:
            template_xsl_rendering = TemplateXslRendering(template=template_id, list_xslt=list_xslt,
                                                          detail_xslt=detail_xslt)
            _upsert(template_xsl_rendering)


def _upsert(template_xsl_rendering):
    """ Upsert an TemplateXslRendering.

    Args:
        template_xsl_rendering: TemplateXslRendering instance.

    Returns:
        TemplateXslRendering instance

    """
    return template_xsl_rendering.save()


def delete(template_xsl_rendering):
    """ Delete an TemplateXslRendering.

    Args:
        template_xsl_rendering: TemplateXslRendering to delete.

    """
    template_xsl_rendering.delete()


def get_by_id(template_xsl_rendering_id):
    """ Get an TemplateXslRendering document by its id.

    Args:
        template_xsl_rendering_id: Id.

    Returns:
        TemplateXslRendering object.

    Raises:
        DoesNotExist: The TemplateXslRendering doesn't exist.
        ModelError: Internal error during the process.

    """
    return TemplateXslRendering.get_by_id(template_xsl_rendering_id)


def get_by_template_id(template_id):
    """Get TemplateXslRendering by its template id.

    Args:
        template_id: Template id.

    Returns:
        The TemplateXslRendering instance.

    """
    return TemplateXslRendering.get_by_template_id(template_id)


def get_by_template_hash(template_hash):
    """Get TemplateXslRendering by its template hash.

    Args:
        template_hash: Template hash.

    Returns:
        The TemplateXslRendering instance.

    """
    instance = None
    # Get list of templates with the given hash
    templates = template_api.get_all_by_hash(template_hash)
    # Check if one of the templates has a xslt. Stop when found.
    # FIXME: Check if this process is okay (no solutions to distinguish templates from same hash)
    for template in templates:
        try:
            instance = TemplateXslRendering.get_by_template_id(template.id)
            break
        except exceptions.DoesNotExist:
            pass

    if instance is None:
        raise exceptions.DoesNotExist("No TemplateXslRendering found with the given template hash")

    return instance


def get_all():
    """Get all TemplateXslRendering.

    Returns:
        List of TemplateXslRendering.

    """
    return TemplateXslRendering.get_all()



""" Fixtures files for Data
"""
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface
from core_main_app.components.data.models import Data
from core_main_app.components.template.models import Template
from core_main_app.components.workspace.models import Workspace


class DataFixtures(FixtureInterface):
    """ Data fixtures
    """
    data_1 = None
    data_2 = None
    template = None
    data_collection = None

    def insert_data(self):
        """ Insert a set of Data.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template()
        self.generate_data_collection()

    def generate_data_collection(self):
        """ Generate a Data collection.

        Returns:

        """
        # NOTE: no xml_content to avoid using unsupported GridFS mock
        self.data_1 = Data(template=self.template, user_id='1', dict_content=None,
                           title='title').save()
        self.data_2 = Data(template=self.template, user_id='2', dict_content=None,
                           title='title2').save()
        self.data_collection = [self.data_1, self.data_2]

    def generate_template(self):
        """ Generate an unique Template.

        Returns:

        """
        template = Template()
        xsd = '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">' \
              '<xs:element name="tag"></xs:element></xs:schema>'
        template.content = xsd
        template.hash = ""
        template.filename = "filename"
        self.template = template.save()


class QueryDataFixtures(DataFixtures):
    """ Data fixtures
    """

    def generate_data_collection(self):
        """ Generate a Data collection.

        Returns:

        """
        content_1 = {
            "root": {
                "element": "value",
                "complex": {
                    "child1": "test",
                    "child2": 0
                }
            }
        }
        content_2 = {
            "root": {
                "element": "value2"
            }
        }
        # NOTE: no xml_content to avoid using unsupported GridFS mock
        self.data_1 = Data(template=self.template, user_id='1', dict_content=content_1,
                           title='title').save()
        self.data_2 = Data(template=self.template, user_id='2', dict_content=content_2,
                           title='title2').save()
        self.data_collection = [self.data_1, self.data_2]


class AccessControlDataFixture(FixtureInterface):
    """ Access Control Data fixture
    """
    USER_1_NO_WORKSPACE = 0
    USER_2_NO_WORKSPACE = 1
    USER_1_WORKSPACE_1 = 2
    USER_2_WORKSPACE_2 = 3

    template = None
    workspace_1 = None
    workspace_2 = None
    data_collection = None

    def insert_data(self):
        """ Insert a set of Data.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template()
        self.generate_workspace()
        self.generate_data_collection()

    def generate_data_collection(self):
        """ Generate a Data collection.

        Returns:

        """
        data_1 = Data(template=self.template,
                      title='Data 1',
                      user_id='1').save()
        data_2 = Data(template=self.template,
                      title='Data 2',
                      user_id='2').save()
        data_3 = Data(template=self.template,
                      title='Data 3',
                      user_id='1',
                      workspace=self.workspace_1.id).save()
        data_4 = Data(template=self.template,
                      title='Data 4',
                      user_id='2',
                      workspace=self.workspace_2.id).save()
        self.data_collection = [data_1, data_2, data_3, data_4]

    def generate_template(self):
        """ Generate an unique Template.

        Returns:

        """
        template = Template()
        xsd = '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">' \
              '<xs:element name="tag"></xs:element></xs:schema>'
        template.content = xsd
        template.hash = ""
        template.filename = "filename"
        self.template = template.save()

    def generate_workspace(self):
        """ Generate an unique Template.

        Returns:

        """
        self.workspace_1 = Workspace(title="Workspace 1",
                                     owner='1',
                                     read_perm_id='1',
                                     write_perm_id='1').save()
        self.workspace_2 = Workspace(title="Workspace 2",
                                     owner='2',
                                     read_perm_id='2',
                                     write_perm_id='2').save()



""" Utils for template version manager rest Apis
"""
import json

from core_main_app.commons.exceptions import RestApiError
from core_main_app.utils.access_control.exceptions import AccessControlError


def load_dependencies(validated_data):
    """Return dependencies as a dict.

    Args:
        validated_data:

    Returns:

    """
    if 'dependencies_dict' in validated_data:
        dependencies_dict = validated_data['dependencies_dict']
        try:
            return json.loads(dependencies_dict)
        except:
            raise RestApiError('Incorrect format of the dependencies parameter.')
    return None


def can_user_modify_template_version_manager(template_version_manager, user):
    """ Check that user can modify the template version manager.

    Args:
        template_version_manager:
        user:

    Returns:

    """
    if user.is_superuser is False and user.is_staff is False:
        if template_version_manager.user != user.id:
            raise AccessControlError("You don't have the permission to update this object.")



"""
Template models
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors
from core_main_app.commons import exceptions
from core_main_app.commons.regex import NOT_EMPTY_OR_WHITESPACES


class Template(Document):
    """Represents an XML schema template that defines the structure of data"""
    filename = fields.StringField(regex=NOT_EMPTY_OR_WHITESPACES)
    content = fields.StringField()
    hash = fields.StringField()
    _display_name = fields.StringField(blank=True)
    dependencies = fields.ListField(fields.ReferenceField("self"), default=[], blank=True)

    meta = {'allow_inheritance': True}

    @staticmethod
    def get_all(is_cls):
        """Return all templates.

        Returns:

        """
        if is_cls:
            # will return all Template object only
            return Template.objects(_cls=Template.__name__).all()
        else:
            # will return all inherited object
            return Template.object().all()

    @staticmethod
    def get_by_id(template_id):
        """Return a template by its id.

        Args:
            template_id:

        Returns:

        """
        try:
            return Template.objects().get(pk=str(template_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as e:
            raise exceptions.ModelError(e.message)

    @staticmethod
    def get_all_by_hash(template_hash):
        """ Return all template having the given hash.

        Args:
            template_hash: Template hash.

        Returns:
            List of Template instance.

        """
        return Template.objects(hash=template_hash).all()

    @staticmethod
    def get_all_by_hash_list(template_hash_list):
        """ Return all template having the given hash list.

        Args:
            template_hash_list: Template hash list.

        Returns:
            List of Template instance.

        """
        return Template.objects(hash__in=template_hash_list).all()

    @staticmethod
    def get_all_by_id_list(template_id_list):
        """ Return all template with id in list.

        Args:
            template_id_list:

        Returns:

        """
        return Template.objects(pk__in=template_id_list)

    @staticmethod
    def get_all_templates_by_dependencies(dependency_id_list):
        """List all templates having the given dependencies.

        Args:
            dependency_id_list: List of dependency ids.

        Returns:
            List of templates.

        """

        return Template.objects(dependencies__in=dependency_id_list).all()

    @property
    def display_name(self):
        """Return template name to display.

        Returns:

        """
        if self._display_name is not None:
            return self._display_name
        else:
            return self.filename

    @display_name.setter
    def display_name(self, value):
        """Set template name to display.

        Args:
            value:

        Returns:

        """
        self._display_name = value



"""Pagination configuration for rest_framework
"""

from rest_framework.pagination import PageNumberPagination
from core_main_app.settings import RESULTS_PER_PAGE


class StandardResultsSetPagination(PageNumberPagination):
    page_size = RESULTS_PER_PAGE



""" Unit Test Data
"""
from core_main_app.utils.integration_tests.integration_base_test_case import MongoIntegrationBaseTestCase
from tests.components.data.fixtures.fixtures import DataFixtures
from core_main_app.components.data.models import Data
from core_main_app.commons import exceptions
from bson.objectid import ObjectId

fixture_data = DataFixtures()


class TestDataGetAll(MongoIntegrationBaseTestCase):

    fixture = fixture_data

    def test_data_get_all_return_collection_of_data(self):
        # Act
        result = Data.get_all()
        # Assert
        self.assertTrue(all(isinstance(item, Data) for item in result))

    def test_data_get_all_return_objects_data_in_collection(self):
        # Act
        result = Data.get_all()
        # Assert
        self.assertTrue(len(self.fixture.data_collection) == result.count())


class TestDataGetById(MongoIntegrationBaseTestCase):

    fixture = fixture_data

    def test_data_get_by_id_raises_api_error_if_not_found(self):
        # Act # Assert
        with self.assertRaises(exceptions.DoesNotExist):
            Data.get_by_id(ObjectId())

    def test_data_get_by_id_return_data_if_found(self):
        # Act
        result = Data.get_by_id(self.fixture.data_1.id)
        # Assert
        self.assertEqual(result, self.fixture.data_1)


class TestDataGetAllByUserId(MongoIntegrationBaseTestCase):

    fixture = fixture_data

    def test_data_get_all_by_user_id_return_collection_of_data_from_user(self):
        # Arrange
        user_id = 1
        # Act
        result = Data.get_all_by_user_id(user_id)
        # Assert
        self.assertTrue(all(item.user_id == str(user_id) for item in result))

    def test_data_get_all_by_user_id_return_empty_collection_of_data_from_user_does_not_exist(self):
        # Arrange
        user_id = 800
        # Act
        result = Data.get_all_by_user_id(user_id)
        # Assert
        self.assertTrue(result.count() == 0)


class TestDataGetAllExceptUserId(MongoIntegrationBaseTestCase):

    fixture = fixture_data

    def test_data_get_all_except_user_id_return_collection_of_data_where_user_is_not_owner(self):
        # Arrange
        user_id = 1
        # Act
        result = Data.get_all_except_user_id(user_id)
        # Assert
        self.assertTrue(item.user_id != user_id for item in result)

    def test_data_get_all_by_user_id_return_full_collection_of_data_from_user_does_not_exist(self):
        # Arrange
        user_id = 800
        # Act
        result = Data.get_all_except_user_id(user_id)
        # Assert
        self.assertTrue(result.count() > 0)



"""BLOB component utils
"""
from django.core.urlresolvers import reverse


def get_blob_download_uri(blob, request):
    """ Return download url for the blob.

    Args:
        blob:
        request:

    Returns:

    """
    # get URI to download blob
    blob_download_uri = request.build_absolute_uri(reverse("core_main_app_rest_blob_download",
                                                           kwargs={'pk': str(blob.id)}))
    return blob_download_uri
